/*contains styles that are always present.*/
grid.board {
  border: 2px solid #333;
  background: #333;
}
grid.block {
  background: #999;
}
sudokucell {
  background: @view_bg_color;
}
label {
  color: @view_fg_color;
}
sudokucell.selected {
  background: shade(@accent_bg_color, 1.3);
}
sudokucell.highlight-coord {
  background: #EDEDED;
}
sudokucell.highlight-number {
  background: shade(@accent_bg_color, 1.5);
}
label.earmark.highlight-number {
  /* reduce red to diffentiate from warnings */
  color: oklab(from @accent_color l min(a, 0.1) b);
  font-weight: bold;
}
label.error {
  color: @error_color;
  font-weight: bold;
}
sudokucell.fixed {
  background: #CCC;
}
overlay.paused > label {
  background: rgba(0, 0, 0, 0.75);
  font-weight: bold;
  color: #FFF;
}
sudokucell.paused {
  background: #FFF;
}
sudokucell.paused > label {
  color: transparent;
}
