/*contains styles only used when AdwStyleManager:dark is TRUE.*/
:root {
  --sudoku-dark-grey: #747474;
}
grid.board {
  border: none;
  background: var(--sudoku-dark-grey);
}
grid.block {
  background: shade(var(--sudoku-dark-grey), 0.79);
}
sudokucell {
  background: shade(var(--sudoku-dark-grey), 0.59);
}
sudokucell.highlight-coord {
  background: shade(var(--sudoku-dark-grey), 0.72);
}
sudokucell.selected {
  background: shade(@accent_bg_color, 0.7);
}
sudokucell.fixed {
  background: shade(var(--sudoku-dark-grey), 0.45);
}
sudokucell.highlight-number {
  background: shade(@accent_bg_color, 0.5);
}
sudokucell.paused {
  background: shade(var(--sudoku-dark-grey), 0.59);
}
